﻿/*
http://www.sno.phy.queensu.ca/~phil/exiftool/
*/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Text;
using System.Threading;
using System.Linq;

namespace BBCSharp
{
    public sealed class ExifToolWrapper : IDisposable
    {
        public string Exe { get; }
        private const string ExeName = "exiftool(-k).exe";
        private const string Arguments       = "-fast  -m -q -q -stay_open True -@ - -common_args -d \"%Y.%m.%d %H:%M:%S\" -c \"%d %d %.6f\" -t";   //-g for groups
        private const string ArgumentsFaster = "-fast2 -m -q -q -stay_open True -@ - -common_args -d \"%Y.%m.%d %H:%M:%S\" -c \"%d %d %.6f\" -t";
        private const string SuccessMessage  = "1 image files updated";
        private const string ExitMessage     = "-- press RETURN --";

        //-fast2 also causes exiftool to avoid extracting any EXIF MakerNote information

        public enum Statuses { Stopped, Starting, Ready, Stopping };
        public Statuses Status { get; private set; }

        public bool Resurrect { get; set; } = true;
        private bool stopRequested = false;

        private int cmdCnt = 0;
        private readonly StringBuilder output = new StringBuilder();
        private readonly ProcessStartInfo psi;
        private Process proc = null;
        private readonly ManualResetEvent waitHandle = new ManualResetEvent(true);

        public string ExiftoolVersion { get; private set; }
        public string LastExiftoolResponse { get; private set; }

        public bool CaptureErrors { get; set; } = false;
        private readonly StringBuilder lastError = new StringBuilder();
        public string LastError => lastError.ToString();

        public ExifToolWrapper(string path = null, bool faster = false)
        {
            // ReSharper disable once AssignNullToNotNullAttribute
            Exe = string.IsNullOrEmpty(path) ? Path.Combine(Path.GetDirectoryName(Assembly.GetEntryAssembly().Location), ExeName) : path;
            if(!File.Exists(Exe))
                throw new ExifToolException(ExeName + " not found");

            psi = new ProcessStartInfo
            {
                FileName = Exe,
                Arguments = faster ? ArgumentsFaster : Arguments,
                CreateNoWindow = true,
                UseShellExecute = false,
                RedirectStandardOutput = true,
                RedirectStandardError = true,
                RedirectStandardInput = true
            };

            Status = Statuses.Stopped;
        }

        private void OutputDataReceived(object sender, DataReceivedEventArgs e)
        {
            if(string.IsNullOrEmpty(e.Data))
                return;

            if(Status == Statuses.Starting)
            {
                ExiftoolVersion = e.Data;
                waitHandle.Set();
            }
            else
            {
                if(string.Equals(e.Data, $"{{ready{cmdCnt}}}", StringComparison.OrdinalIgnoreCase))
                    waitHandle.Set();
                else
                    output.AppendLine(e.Data);
            }
        }

        private void ErrorDataReceived(object sender, DataReceivedEventArgs e)
        {
            if(string.IsNullOrEmpty(e?.Data))
                return;

            if(string.Equals(e.Data, ExitMessage, StringComparison.OrdinalIgnoreCase))
                proc?.StandardInput.WriteLine();
            else if(CaptureErrors)
                lastError.AppendLine(e.Data);
        }

        public void Start()
        {
            stopRequested = false;

            if(Status != Statuses.Stopped)
                throw new ExifToolException("Process is not stopped");

            Status = Statuses.Starting;

            proc = new Process { StartInfo = psi, EnableRaisingEvents = true };
            proc.OutputDataReceived += OutputDataReceived;
            proc.ErrorDataReceived += ErrorDataReceived;
            proc.Exited += proc_Exited;
            proc.Start();

            proc.BeginOutputReadLine();
            proc.BeginErrorReadLine();
            proc.StandardInput.AutoFlush = true;

            waitHandle.Reset();
            proc.StandardInput.Write("-ver\n-execute0000\n");
            waitHandle.WaitOne();

            Status = Statuses.Ready;
        }

        //detect if process was killed
        void proc_Exited(object sender, EventArgs e)
        {
            if(proc != null)
            {
                proc.Dispose();
                proc = null;
            }

            Status = Statuses.Stopped;

            waitHandle.Set();

            if(!stopRequested && Resurrect)
                Start();
        }

        public void Stop()
        {
            stopRequested = true;

            if(Status != Statuses.Ready)
                throw new ExifToolException("Process must be ready");

            Status = Statuses.Stopping;

            waitHandle.Reset();
            proc.StandardInput.Write("-stay_open\nFalse\n");
            
            if(!waitHandle.WaitOne(5000))
            {
                if(proc != null)
                {
                    try
                    {
                        proc.Kill();
                        proc.WaitForExit(2000);
                        proc.Dispose();
                    }
                    catch(Exception xcp)
                    {
                        Debug.WriteLine(xcp.ToString());
                    }

                    proc = null;
                }

                Status = Statuses.Stopped;
            }
        }

        private static readonly object lockObj = new object();

        public string SendCommand(string cmd, params object[] args)
        {
            if(Status != Statuses.Ready)
                throw new ExifToolException("Process must be ready");

            string resp;
            lock(lockObj)
            {
                if(CaptureErrors)
                    lastError.Clear();

                waitHandle.Reset();
                proc.StandardInput.Write("{0}\n-execute{1}\n", args.Length == 0 ? cmd : string.Format(cmd, args), cmdCnt);
                waitHandle.WaitOne();
                cmdCnt++;

                resp = LastExiftoolResponse = output.ToString();
                output.Clear();
            }

            return resp;
        }

        public bool SetExifInto(string path, string key, string val, bool overwriteOriginal = true)
        {
            return SetExifInto(path, new Dictionary<string, string> {{key, val}}, overwriteOriginal);
        }

        public bool SetExifInto(string path, Dictionary<string, string> data, bool overwriteOriginal = true)
        {
            if(!File.Exists(path))
                return false;

            StringBuilder cmd = new StringBuilder();
            foreach(KeyValuePair<string, string> kv in data)
            {
                cmd.AppendFormat("-{0}={1}\n", kv.Key, kv.Value);
            }
            
            if(overwriteOriginal)
                cmd.Append("-overwrite_original\n");
            
            cmd.Append(path);
            string sRes = SendCommand(cmd.ToString());

            return sRes.Contains(SuccessMessage);
        }

        public Dictionary<string, string> FetchExifFrom(string path, IEnumerable<string> tagsToKeep = null, bool keepKeysWithEmptyValues = true)
        {
            Dictionary<string, string> res = new Dictionary<string, string>();

            var tagsTable = tagsToKeep?.ToDictionary(x => x, x => 1);
            bool filter = tagsTable != null && tagsTable.Count > 0;
            string sRes = SendCommand(path);
            foreach(string s in sRes.Split(new [] { Environment.NewLine }, StringSplitOptions.RemoveEmptyEntries))
            {
                string[] kv = s.Split('\t');
                Debug.Assert(kv.Length == 2, "Can not parse line :'" + s + "'");

                if(kv.Length != 2 || (!keepKeysWithEmptyValues && string.IsNullOrEmpty(kv[1])))
                    continue;

                if(filter && !tagsTable.ContainsKey(kv[0]))
                    continue;
                    
                res[kv[0]] = kv[1];
            }

            return res;
        }

        public List<string> FetchExifToListFrom(string path, IEnumerable<string> tagsToKeep = null, bool keepKeysWithEmptyValues = true, string separator = ": ")
        {
            List<string> res = new List<string>();

            var tagsTable = tagsToKeep?.ToDictionary(x => x, x => 1);
            bool filter = tagsTable != null && tagsTable.Count > 0;
            string sRes = SendCommand(path);
            foreach(string s in sRes.Split(new[] { Environment.NewLine }, StringSplitOptions.RemoveEmptyEntries))
            {
                string[] kv = s.Split('\t');
                Debug.Assert(kv.Length == 2, "Can not parse line :'" + s + "'");

                if(kv.Length != 2 || (!keepKeysWithEmptyValues && string.IsNullOrEmpty(kv[1])))
                    continue;

                if(filter && !tagsTable.ContainsKey(kv[0]))
                    continue;

                res.Add($"{kv[0]}{separator}{kv[1]}");
            }

            return res;
        }

        public bool CloneExif(string source, string dest, bool backup = false)
        {
            if(!File.Exists(source) || !File.Exists(dest))
                return false;

            string sRes = SendCommand("{0}-tagsFromFile\n{1}\n{2}", backup ? "" : "-overwrite_original\n", source, dest);

            return sRes.Contains(SuccessMessage);
        }

        public bool ClearExif(string source, bool backup = false)
        {
            if(!File.Exists(source))
                return false;

            string sRes = SendCommand("{0}-all=\n{1}", backup ? "" : "-overwrite_original\n", source);

            return sRes.Contains(SuccessMessage);
        }

        public DateTime? GetCreationTime(string path)
        {
            DateTime dt;
            if(DateTime.TryParseExact(SendCommand("-DateTimeOriginal\n-s3\n{0}", path), 
                "yyyy.MM.dd HH:mm:ss", 
                CultureInfo.InvariantCulture,
                DateTimeStyles.AllowWhiteSpaces, 
                out dt))
                return dt;

            return null;
        }

        public int GetOrientation(string path)
        {
            int o;
            if(int.TryParse(SendCommand("-Orientation\n-n\n-s3\n{0}", path).Trim('\t', '\r', '\n'), out o))
                return o;

            return 1;
        }

        public int GetOrientationDeg(string path)
        {
            return OrientationPos2Deg(GetOrientation(path));
        }

        public bool SetOrientation(string path, int ori, bool overwriteOriginal = true)
        {
            if(!File.Exists(path))
                return false;

            StringBuilder cmd = new StringBuilder();
            cmd.AppendFormat("-Orientation={0}\n-n\n-s3\n", ori);

            if(overwriteOriginal)
                cmd.Append("-overwrite_original\n");

            cmd.Append(path);
            string sRes = SendCommand(cmd.ToString());

            return sRes.Contains(SuccessMessage);
        }

        public bool SetOrientationDeg(string path, int ori, bool overwriteOriginal = true)
        {
            return SetOrientation(path, OrientationDeg2Pos(ori), overwriteOriginal);
        }

        #region Static orientation helpers

        /*
         
1        2       3      4         5            6           7          8

888888  888888      88  88      8888888888          88  8888888888   88          
88          88      88  88      88  88          88  88      88  88   88  88      
8888      8888    8888  8888    88          8888888888          88   8888888888  
88          88      88  88
88          88  888888  888888

        1 => 'Horizontal (normal)',
        2 => 'Mirror horizontal',
        3 => 'Rotate 180',
        4 => 'Mirror vertical',
        5 => 'Mirror horizontal and rotate 270 CW',
        6 => 'Rotate 90 CW',
        7 => 'Mirror horizontal and rotate 90 CW',
        8 => 'Rotate 270 CW'
         */

        public static int OrientationPos2Deg(int pos)
        {
            switch(pos)
            {
                case 8:
                    return 270;
                case 3:
                    return 180;
                case 6:
                    return 90;
                default:
                    return 0;
            }
        }

        public static int OrientationDeg2Pos(int deg)
        {
            switch(deg)
            {
                case 270:
                    return 8;
                case 180:
                    return 3;
                case 90:
                    return 6;
                default:
                    return 1;
            }
        }

        public static int OrientationString2Deg(string pos)
        {
            switch(pos)
            {
                case "Rotate 270 CW":
                    return 270;
                case "Rotate 180":
                    return 180;
                case "Rotate 90 CW":
                    return 90;
                default:
                    return 0;
            }
        }

        public static string OrientationDeg2String(int deg)
        {
            switch(deg)
            {
                case 270:
                    return "Rotate 270 CW";
                case 180:
                    return "Rotate 180";
                case 90:
                    return "Rotate 90 CW";
                default:
                    return "Horizontal (normal)";
            }
        }

        private static readonly int[] orientationPositions = { 1, 6, 3, 8 };

        public static int RotateOrientation(int crtOri, bool clockwise, int steps = 1)
        {
            int newOri = 1;
            int len = orientationPositions.Length;

            if(steps % len == 0)
                return crtOri;

            for(int i = 0; i < len; i++)
            {
                if(crtOri == orientationPositions[i])
                {
                    newOri = clockwise
                        ? orientationPositions[(i + steps) % len]
                        : orientationPositions[(i + (1 + steps / len) * len - steps) % orientationPositions.Length];

                    break;
                }
            }

            return newOri;
        }

        #endregion
        
        #region IDisposable Members

        public void Dispose()
        {
            Debug.Assert(Status == Statuses.Ready || Status == Statuses.Stopped, "Invalid state");

            if(proc != null && Status == Statuses.Ready)
                Stop();

            waitHandle.Dispose();
        }

        #endregion
    }
}
